/*******************************************************************************
 *
 * e2part.c
 *
 * part of ext2 NT tools
 *
 * Copyright (C) Peeter Joot, 1997
 *
 * This file may be redistributed under the terms of the GNU Public License.
 ******************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "e2p/e2p.h"
#include "dev.h"
#include "winperror.h"
#include "getopt.h"
#include <winioctl.h>
#include <largeint.h>

/*******************************************************************************
 *
 * Functions for listing of partition data.
 *
 ******************************************************************************/

static void print_pinfo(HANDLE dev, PARTITION_INFORMATION * p, 
	int dump_blocks, int use_offsets);

typedef struct {
	DWORD PartitionType;
	char * desc;
} partition_info;

#ifdef __GNUC__
// cygnus gcc doesn't have winioctl.h, but this isn't the only thing to be fixed
// to get this program to compile by gcc

#define PARTITION_ENTRY_UNUSED		0x00	// Entry unused
#define PARTITION_FAT_12		0x01	// 12-bit FAT entries
#define PARTITION_XENIX_1		0x02	// Xenix root
#define PARTITION_XENIX_2		0x03	// Xenix usr
#define PARTITION_FAT_16		0x04	// 16-bit FAT entries
#define PARTITION_EXTENDED		0x05	// Extended partition entry
#define PARTITION_HUGE			0x06	// Huge partition MS-DOS V4
#define PARTITION_IFS			0x07	// IFS Partition: HPFS, NTFS
#define PARTITION_FAT32			0x0B	// FAT32
#define PARTITION_FAT32_XINT13		0x0C	// FAT32 using extended int13 services
#define PARTITION_XINT13		0x0E	// Win95 partition using extended int13 services
#define PARTITION_XINT13_EXTENDED	0x0F	// Same as type 5 but uses extended int13 services
#define PARTITION_PREP			0x41	// PowerPC Reference Platform (PReP) Boot Partition
#define PARTITION_UNIX			0x63	// Unix, GNU HURD
#endif

#define PARTITION_AIX			0x08	// AIX
#define PARTITION_AIX_BOOT		0x09	// AIX Bootable
#define PARTITION_OS2_BOOT_MANAGER	0x0A	// OS/2 Boot Manager
#define PARTITION_VENIX			0x40	// Venix 80286
#define PARTITION_NOVELL1		0x51	// Novell?
#define PARTITION_MICROPORT		0x52	// Microport
#define PARTITION_NOVELLN1		0x64	// Novell Netware
#define PARTITION_NOVELLN2		0x65	// Novell Netware
#define PARTITION_PCIX			0x75	// PC/IX
#define PARTITION_OLDMINIX		0x80	// Old MINIX
#define PARTITION_LINUX_MINUX		0x81	// Linux/MINIX
#define PARTITION_LINUX_SWAP		0x82	// Linux swap
#define PARTITION_LINUX_NATIVE		0x83	// Linux native
#define PARTITION_AMOEBA		0x93	// Amoeba
#define PARTITION_AMOEBA_BBT		0x94	// Amoeba BBT
#define PARTITION_BSD_386		0xA5	// BSD/386
#define PARTITION_BSDI			0xB7	// BSDI fs
#define PARTITION_BSDI_SWAP		0xB8	// BSDI swap
#define PARTITION_SYRINX		0xC7	// Syrinx
#define PARTITION_CPM			0xDB	// CP/M
#define PARTITION_DOS_ACCESS		0xE1	// DOS access
#define PARTITION_DOS_RO		0xE3	// DOS R/O
#define PARTITION_DOS_SECONDARY		0xF2	// DOS secondary
#define PARTITION_BBT			0xFF	// BBT
#define PARTITION_UNRECOGNIZED		0xFFFFFFFF

/* probably won't ever have more than 16 partitions on a drive. */
#define MAX_PARTITIONS 16

/* structure defintion for IOCTL_DISK_GET_DRIVE_LAYOUT query */
typedef struct {
        DWORD PartitionCount;
        DWORD Signature;
        PARTITION_INFORMATION  PartitionEntry[MAX_PARTITIONS];
} pinfo;

static const partition_info pdesc[] = {
	 { PARTITION_ENTRY_UNUSED, "Entry unused"}
	,{ PARTITION_FAT_12, "12-bit FAT entries"}
	,{ PARTITION_XENIX_1, "Xenix root"}
	,{ PARTITION_XENIX_2, "Xenix usr"}
	,{ PARTITION_FAT_16, "16-bit FAT entries"}
	,{ PARTITION_EXTENDED, "Extended partition entry"}
	,{ PARTITION_HUGE, "Huge partition MS-DOS V4"}
	,{ PARTITION_IFS, "IFS Partition: NTFS, HPFS"}
	,{ PARTITION_AIX, "AIX" }
	,{ PARTITION_AIX_BOOT, "AIX Bootable" }
	,{ PARTITION_OS2_BOOT_MANAGER, "OS/2 Boot Manager"}
	,{ PARTITION_FAT32, "FAT32"}
	,{ PARTITION_FAT32_XINT13, "FAT32 using extended int13 services"}
	,{ PARTITION_XINT13, "Win95 partition using extended int13 services"}
	,{ PARTITION_XINT13_EXTENDED, "Same as PARTITION_EXTENDED but uses extended int13 services"}
	,{ PARTITION_PREP, "PowerPC Reference Platform (PReP) Boot Partition"}
	,{ PARTITION_VENIX, "Venix 80286" }
	,{ PARTITION_NOVELL1, "Novell?" }
	,{ PARTITION_MICROPORT, "Microport" }
	,{ PARTITION_UNIX, "Unix, GNU HURD"}
	,{ PARTITION_NOVELLN1, "Novell Netware" }
	,{ PARTITION_NOVELLN2, "Novell Netware" }
	,{ PARTITION_PCIX, "PC/IX" }
	,{ PARTITION_OLDMINIX, "Old MINIX" }
	,{ PARTITION_LINUX_MINUX, "Linux/MINIX" }
	,{ PARTITION_LINUX_NATIVE, "Linux Native" }
	,{ PARTITION_LINUX_SWAP, "Linux Swap"}
	,{ PARTITION_AMOEBA, "Amoeba" }
	,{ PARTITION_AMOEBA_BBT, "Amoeba BBT" }
	,{ PARTITION_BSD_386, "BSD/386" }
	,{ PARTITION_BSDI, "BSDI fs" }
	,{ PARTITION_BSDI_SWAP, "BSDI swap" }
	,{ PARTITION_SYRINX, "Syrinx" }
	,{ PARTITION_CPM, "CP/M" }
	,{ PARTITION_DOS_ACCESS, "DOS access" }
	,{ PARTITION_DOS_RO, "DOS R/O" }
	,{ PARTITION_DOS_SECONDARY, "DOS secondary" }
	,{ PARTITION_BBT, "BBT" }
	,{ PARTITION_UNRECOGNIZED, "unrecognized partition type" }
	};

int sb_print = 0;
int use_floppy = 0;

static void hexdump(char buf[1024]);
static int partition_dump(HANDLE dev, LARGE_INTEGER offset, unsigned blocks);

/* see in raw form what is on the disk: */
static int partition_dump(HANDLE dev, LARGE_INTEGER offset, unsigned blocks)
{
	char buf[1024];
	unsigned i;
	LARGE_INTEGER off;

	for ( i = 0 ; i < blocks ; i++ ) {
		int rc;
		off = LargeIntegerAdd(ConvertUlongToLargeInteger(1024*i),offset);
		if ( (rc = nt_raw_read(dev, off, 1024, buf) ))
			return rc;
			
		hexdump(buf);
	}
	return 0;
}

/* hex dump of a block of 1024 characters */
static void hexdump(char buf[1024])
{
	int i;
	for ( i = 0 ; i < 1024; ) {
		int j = i;
		for ( ; j < i + 24 ; j++ ) {
			if ( j < 1024 ) {
				int c = buf[j]&0xff;
				printf("%02x",c);
			}
			else
				printf("  ");
		}
		putchar(' ');
		j = i;
		for ( ; j < i + 24 ; j++ ) {
			if ( j < 1024 ) {
				int c = buf[j]&0xff;
				if ( isprint(c) ) 
					printf("%c",c);
				else 
					putchar('.');
			}
			else
				break;
		}
		i += 24;
		putchar('\n');
	}
}

int single_partitiondump(unsigned drivenum, unsigned partition, 
	unsigned dump_blocks)
{
	HANDLE dev;
	char dnamebuf[64];
	char dname[64];
	char dnamel[64];
	PARTITION_INFORMATION pbuf;
	DWORD bret;
	int rc = 0;

	sprintf(dname,"dev_hd%c%i", 'a' + (char)drivenum, partition); 
	sprintf(dnamel,"\\Device\\Harddisk%i\\Partition%i", drivenum, partition); 
	if ( QueryDosDevice(dname, dnamebuf, sizeof(dnamebuf)) == 0) {
		rc = GetLastError();
//		printf("query errcode = %i\n", rc);
		if ( rc == ERROR_FILE_NOT_FOUND ) {
			if ( DefineDosDevice(DDD_RAW_TARGET_PATH, 
				dname, dnamel) == 0 ) 
			return winperror("couldn't define device: %s", dnamel);
		}
		else 
			return winperror("query dos device");
	}

	sprintf(dnamebuf,"\\\\.\\%s", dname);

	printf("accessing Device: \"%s\" as DosDevice \"%s\"\n"
		"via file \"%s\"\n", dnamel, dname, dnamebuf );

	dev = CreateFile(
		dnamebuf
		, GENERIC_READ
		, FILE_SHARE_READ
#if 0
		| FILE_SHARE_WRITE
#endif
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_READONLY
		/* | FILE_FLAG_RANDOM_ACCESS */
		/* | FILE_FLAG_OVERLAPPED // asycronous i/o */
		, NULL
		);

	if ( dev == INVALID_HANDLE_VALUE )
		return winperror("CreateFile failure");

        if ( DeviceIoControl(
                dev
                , IOCTL_DISK_GET_PARTITION_INFO 
		, NULL
                , 0
                , &pbuf
                , sizeof(PARTITION_INFORMATION)
                , &bret
                , NULL
                ) == FALSE ) {
                        rc = winperror("DeviceIoControl failure");
			goto cleanup;
                }

	print_pinfo(dev, &pbuf, dump_blocks, 0);

cleanup:
	if ( CloseHandle(dev) == 0 )
		return winperror("CloseHandle failure");

        return rc;			
}

#define ONE_MB	(1024*1024)

static void print_pinfo(HANDLE dev, PARTITION_INFORMATION * p, 
	int dump_blocks, int use_offset)
{
	char * s;	
	int j;
	struct ext2_super_block sb;
	unsigned type;
	LARGE_INTEGER offset;
	LARGE_INTEGER off;

	printf(
		"starting offset: 0x%I64x\n"
		"partition length: 0x%I64x (%lu Mb)\n"
		"hiddensectors: 0x%x\n"
		"partition number: %x\n"
		"partition type: 0x%x\n"
		"bootindicator: %x\n"
		"recognized partition: %x\n"
		"rewritepartition: %x\n"
		,p->StartingOffset
		,p->PartitionLength
		,p->PartitionLength.LowPart / ONE_MB
		,p->HiddenSectors
		,p->PartitionNumber
		,p->PartitionType
		,p->BootIndicator
		,p->RecognizedPartition
		,p->RewritePartition
		);

	type = p->PartitionType;
	s = "un";
	if ( IsRecognizedPartition(type) ) s = "";
	printf("%srecognized (by windows) partition type = 0x%x\n",s,type);
	
	for ( j = 0 ; j < sizeof(pdesc)/sizeof(partition_info) -1; j++ )
		if ( type == pdesc[j].PartitionType ) break;

	printf("partition type is : %s\n", pdesc[j].desc);
	if ( type == PARTITION_LINUX_NATIVE ) {
		if ( use_offset ) {
			offset = p->StartingOffset;
			off = LargeIntegerAdd(ConvertUlongToLargeInteger(1024),offset);
		}
		else {
			off = ConvertUlongToLargeInteger(1024);
			offset = ConvertUlongToLargeInteger(0);
		}
		if ( check_super(dev,off,&sb) ) 
			return;

	if ( sb_print )
		list_super(&sb);

	/* hexdump of the first dump_blocks blocks of the partition */
		if ( dump_blocks )
			partition_dump(dev, offset, dump_blocks);
		}
	printf("\n");
}

//
// GetDiskGeometry
//
BOOL GetDiskGeometry(HANDLE hDisk, PDISK_GEOMETRY lpGeometry) {

	DWORD ReturnedByteCount;

	return DeviceIoControl(hDisk, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0,
                lpGeometry, sizeof(*lpGeometry), &ReturnedByteCount,NULL);
}

int print_drive_geometry = 0;

//
// PrintGeometry
//
VOID PrintGeometry(PDISK_GEOMETRY lpGeometry) {

        printf("Size = %d Mb - ", lpGeometry->Cylinders.LowPart * 
		lpGeometry->SectorsPerTrack * lpGeometry->TracksPerCylinder 
		* 512 / ONE_MB);

	printf("Cyl = %d - Tr/Cyl = %d - Sec/Tr = %d\n", 
		lpGeometry->Cylinders.LowPart, lpGeometry->TracksPerCylinder, 
		lpGeometry->SectorsPerTrack);
}

int partitiondata(unsigned drivenum, HANDLE * dev, pinfo * pbuf)
{
	char dnamebuf[25];
	DWORD bret; /* bytes returned */
	
	/* paranoia check: make sure we don't overrun this string by mistake. */
	if ( drivenum > 99 ) {
		fprintf(stderr,"Bad drive number %i.  Expected something much "
			"smaller.\n",drivenum);
		return drivenum;
	}

	sprintf(dnamebuf,"\\\\.\\PhysicalDrive%i",drivenum);
	
	*dev = CreateFile(
		dnamebuf
		, GENERIC_READ
		, FILE_SHARE_READ
		/* | FILE_SHARE_WRITE */
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_READONLY 
		/* | FILE_FLAG_RANDOM_ACCESS */
		/* | FILE_FLAG_OVERLAPPED // asycronous i/o */
		, NULL
		);

	if ( *dev == INVALID_HANDLE_VALUE ) {
		*dev = NULL;
		return winperror("CreateFile failure");
		}

	if ( print_drive_geometry ) {
		DISK_GEOMETRY Geometry;
		if ( !GetDiskGeometry(*dev,&Geometry) ) 
			winperror("GetDiskGeometry");
		else {
			printf(
"\nPhysical Drive %i specifications"
"\n-------------------------------\n"
			, drivenum);
			PrintGeometry(&Geometry);
			}
		}

	if ( DeviceIoControl(
		*dev
		, IOCTL_DISK_GET_DRIVE_LAYOUT
		, NULL
		, 0
		, pbuf
		, sizeof(pinfo)
		, &bret
		, NULL
		) == FALSE ) {
			int rc = winperror("DeviceIoControl failure");
			if ( CloseHandle(*dev) == 0 )
				winperror("CloseHandle failure");
			*dev = NULL;
			return rc;
		}

	return 0;
}

int list_partitiondata(unsigned drivenum, unsigned dump_blocks)
{
	HANDLE dev;
	pinfo pbuf;
	int rc;
	unsigned i;

	if ( rc = partitiondata(drivenum, &dev, &pbuf) )
		goto cleanup;

	printf(	"\nPartition Table Data"
		"\n--------------------\n");
	for ( i = 0 ; i < pbuf.PartitionCount ; i++ ) {
		if (!pbuf.PartitionEntry[i].PartitionType) continue;
		print_pinfo(dev, &(pbuf.PartitionEntry[i]), dump_blocks, 1);
		}

cleanup:
	if ( dev && CloseHandle(dev) == 0 )
		return winperror("CloseHandle failure");
		
	return rc;
}

void Usage()
{
	fprintf(stderr,
"SYNOPSIS\n"
"\tE2PART [-ghs] [-d dump] diskno [partition]\n"
"\n"
"DESCRIPTION\n"
"\tThe E2PART program lists the partitions that are available on a\n"
"\tparticular physical disk. This program does not use the E2CWD\n"
"\tenvironment variable, instead the disk number is given as an\n"
"\targument to the E2PART program.\n"
"\n"
"\tdiskno is a integer in the open range 0, 1, 2, ...\n"
"\tdiskno specifies the physical disk to examine and has a value\n"
"\tgreater than or equal to 0.\n"
"\n"
"\tpartition can be used with disk to check for a specific ext2\n"
"\tpartition.  Partition numbers start at 1.\n"
"\n"
"-d dump\tDo a hex dump of dump 1024 byte blocks of the ext2 partition.\n"
"\tThis is for primarily for debugging.\n"
"\n"
"-s\tPrint out the superblock of any ext2 partitions.\n"
"\n"
"-g\tPrint out drive geometry.\n"
"\n"
"-h\tPrint this message.\n"
	);
}

int main(int argc, char * argv[]) 
{
	int n, p, c;
	int num_dumped = 0;

	while ( (c = getopt(argc, argv, "gsd:h")) != -1 ) {
		switch ( c ) {
			case 'g' :
				print_drive_geometry = 1;
				break;
			case 'd' :
				if ( sscanf(optarg,"%i",&num_dumped) != 1 ) {
					fprintf(stderr,"couldn't parse %s\n",
						optarg);
					return 2;
					}	
				break;
			case 's' :
				sb_print = 1;
				break;
			case 'F' :
				use_floppy = 1;
				break;
			case 'h' :
			case '?' :
			default :
				Usage();
				return 1;
			}
		}

	if ( optind == argc ) {
		Usage();
		return 4;
		}
		
	if ( sscanf(argv[optind],"%i",&n) != 1 ) {
		fprintf(stderr,"Error: could not parse disk number %s.\n",
			argv[optind]);
		return 2;
		}

	if ( argc == ++optind && !use_floppy ) 
		return list_partitiondata(n, num_dumped);

	if ( sscanf(argv[optind],"%i",&p) != 1 ) {
		fprintf(stderr,"Error: could not parse partition number %s.\n",
			argv[optind]);
		return 3;
		}
	return single_partitiondump(n,p,num_dumped);
}
